#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment drawing procedures

# (c) Dinasset 2021
#
#   History:
#
#   v1.0: 2021-05-31: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
deliwrkwh=5400   
mediwrkwh=3600   
hardwrkwh=2400   
SIter=10    # was 30
MIter=15    # was 40
LIter=20    # was 50
SRadius=1.5     
MRadius=2.5 #1.5     
LRadius=3.5 #2.0
SAmount=0.0     
MAmount=0.0   #0.015     
LAmount=0.0   #0.020      

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()

def gegl_c2g(image, drawable, uradius, usamples, uiter):
    radius=int(uradius)      #dft 300   #ui_range (2, 1000)
    samples=int(usamples)    #dft 4     #ui_range (3, 17)
    iterations=int(uiter)    #dft 10    #ui_range (1, 30)
    enhance_shadows=False
    mystring="gegl:c2g radius=%d samples=%d iterations=%d enhance-shadows=%d"
    gegl_graph_string=mystring % (radius, samples, iterations, enhance_shadows)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_meancurvatureblur (image, drawable, uiter):
    iterations=int(uiter)
    gegl_graph_string="gegl:mean-curvature-blur iterations=%d" % (iterations)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_noise_cie_lch (image, drawable, uhold, ulightd, uchromad, uhued, useed):
    holdness=int(uhold)
    lightness_distance=float(ulightd)
    chroma_distance=float(uchromad)
    hue_distance=float(uhued)
    seed=int(useed)
    mystring="gegl:noise-cie-lch holdness=%d lightness-distance=%f chroma-distance=%f hue-distance=%f seed=%d"
    gegl_graph_string=mystring % (holdness, lightness_distance, chroma_distance, hue_distance, seed)
    gegl_graph(image, drawable,gegl_graph_string)
    
def gegl_saturation (image, drawable, uscale, ucolorspace):  
    scale=uscale                   
    colorspace=ucolorspace                
    mystring="gegl:saturation scale=%f colorspace=%d"
    gegl_graph_string=mystring % (scale, colorspace)
    gegl_graph(image, drawable,gegl_graph_string)
  
# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer, wrkwh):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (work) output image -------------------------
    scale = (wrkwh*1.0)/((owidth+oheight)*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    # adjust mean values
    adjustmeanv (newImage, newBG)
    # remove low values, remove high values  
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                   1.00, 0.10, 0.90, False)
    newBG.opacity = 100.0
    
    # add noise to create strokes look
    Noised = newBG.copy()
    newImage.add_layer (Noised, 0)
    HOLD=2; LIGHTD=40.0; CHROMAD=40.0; HUED=3.0; SEED=0
    # scaling to enlarge noise dots
    pdb.gimp_layer_scale (Noised , rwidth/4, rheight/2, True)
    gegl_noise_cie_lch (newImage, Noised, HOLD, LIGHTD, CHROMAD, HUED, SEED)
    pdb.gimp_layer_scale (Noised , rwidth, rheight, True)
    Noised.opacity = 25.0
    newBG = pdb.gimp_image_merge_down (newImage, Noised, CLIP_TO_IMAGE)
    newBG.opacity = 100.0
    newBG.name = "NewImage: Background Layer from "+oname+" (noised)"
    
    return(newImage)

#---------------------------------------------------------------------------
# adjust values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer):
    # new version modified to take care of the existing mean in determining the theretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    # consider 3 classes
    if meanl < 85: thmean = 100
    elif meanl > 170: thmean = 155
    else: thmean = 127.5    # regular
    minmean = thmean-16
    maxmean = thmean+16
    
    modified=False
    if meanl < minmean:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        while meanl < minmean:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        while meanl > maxmean:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))
    
    return(inLayer)
     
#--------------------------------------------------------------------------- 
    

############################################################################
# Main
#---------------------------------------------------------------------------

def EasyEffectStyleB (inImage, inLayer,
                      widthheight, drawlook, edgeslook, texturelook, 
                      doFlatten):
    
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # display messages using error console
    
    if drawlook==1: # B&W
        option = "B&W (C2G)"
    elif drawlook==0:  # colours
        option = "colours"
    elif drawlook==2:  # mix
        option = "mix colours-B&W (C2G)"
    
    if texturelook==0: # delicate
        option = option+" (delicate texture)"
    elif texturelook==1:  # medium
        option = option+" (normal texture)"
    elif texturelook==2:  # strong
        option = option+" (strong texture)"
        
    if texturelook==0: wrkwh=deliwrkwh
    elif texturelook==1: wrkwh=mediwrkwh
    elif texturelook==2: wrkwh=hardwrkwh
    
    newImage = createOutimage (inImage, inLayer, wrkwh)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width=Preprocessed.width
    height=Preprocessed.height
    
    #pdb.gimp_display_new (newImage)
    #return
    
    if edgeslook==0:  # small
        mcbiter=SIter
    if edgeslook==1:  # medium 
        mcbiter=MIter
    if edgeslook==2:  # large 
        mcbiter=LIter
    # GEGL Median Blur
    Smoothed = Preprocessed.copy()
    newImage.add_layer (Smoothed, 0)
    Smoothed.name = "Smoothed"
    gegl_meancurvatureblur (newImage, Smoothed, mcbiter)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    Edges = Smoothed.copy()
    newImage.add_layer (Edges, 0)
    Edges.name = "Edges"
    if edgeslook==0:  # small
        RADIUS=SRadius
        AMOUNT=SAmount
    if edgeslook==1:  # medium 
        RADIUS=MRadius
        AMOUNT=MAmount
    if edgeslook==2:  # large 
        RADIUS=LRadius
        AMOUNT=LAmount
    pdb.gimp_layer_flatten (Edges)   
    pdb.plug_in_neon (newImage, Edges,
                      RADIUS, AMOUNT)
    pdb.gimp_drawable_curves_spline (Edges, HISTOGRAM_VALUE, 6, (0.0,0.5, 0.125,1.0, 1.0,1.0))
    pdb.plug_in_vinvert (newImage, Edges)
    Edges.mode = LAYER_MODE_GRAIN_MERGE     #MULTIPLY 
    
    #pdb.gimp_display_new (newImage)
    #return
    
    Processed = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible")
    newImage.add_layer (Processed, 0)
    Processed.mode = LAYER_MODE_NORMAL
    Processed.opacity = 100.0
    Processed.name = "Processed using option "+option
     
    Reinforced = Processed.copy()
    newImage.add_layer (Reinforced, 0)
    Reinforced.name = "To reinforce drawing "
    Reinforced.mode = LAYER_MODE_GRAIN_MERGE
    Reinforced.opacity = 70.0
     
    #pdb.gimp_display_new (newImage)
    #return
    
    PreOutcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible")
    newImage.add_layer (PreOutcome, 0)
    PreOutcome.mode = LAYER_MODE_NORMAL
    PreOutcome.opacity = 5.0   # a bit of gray for colour option
    PreOutcome.name = "Preliminary Outcome "
    # scaling to improve c2g performance
    pdb.gimp_layer_scale (PreOutcome, width/2, height/2, True)
    gegl_c2g(newImage, PreOutcome, 1000, 15, 10)                # good but slow
    pdb.gimp_layer_scale (PreOutcome, width, height, True)
    if drawlook==1: # B&W
        PreOutcome.opacity = 100.0
    if drawlook==2: # Mix
        PreOutcome.opacity = 80.0
        
    # check mean value    
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (PreOutcome,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    # check mean value
    if meanl < 64:
        pdb.gimp_drawable_curves_spline (PreOutcome, HISTOGRAM_VALUE,
                                         6, (0.0,0.025, 0.5,0.7, 1.0,1.0))
    elif meanl < 118:
        pdb.gimp_drawable_curves_spline (PreOutcome, HISTOGRAM_VALUE,
                                         6, (0.0,0.0, 0.5,0.6, 1.0,1.0))
    elif meanl > 138:
        pdb.gimp_drawable_curves_spline (PreOutcome, HISTOGRAM_VALUE,
                                         6, (0.0,0.0, 0.6,0.5, 1.0,1.0))
    elif meanl > 192:
        pdb.gimp_drawable_curves_spline (PreOutcome, HISTOGRAM_VALUE,
                                         6, (0.0,0.0, 0.7,0.5, 1.0,0.975))
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # original on top
    OriginalOnTop = Original.copy()
    newImage.add_layer (OriginalOnTop, 0)
    OriginalOnTop.name = "Original on top for color"
    pdb.gimp_item_set_visible (OriginalOnTop, True)
    OriginalOnTop.mode = LAYER_MODE_LCH_COLOR # for color option
    OriginalOnTop.opacity = 100.0  
    
    # process look option
    if drawlook==0: # colour
        OriginalOnTop.opacity = 100.0
    else:  # B&W or mix
        OriginalOnTop.opacity = 50.0
        if drawlook==1: # B&W
            # scaling to improve c2g performance
            pdb.gimp_layer_scale (OriginalOnTop, width/2, height/2, True)
            gegl_c2g(newImage, OriginalOnTop, 1000, 15, 10)                # good but slow
            pdb.gimp_layer_scale (OriginalOnTop, width, height, True)
            pdb.plug_in_unsharp_mask (newImage, OriginalOnTop, 3.0, 0.5, 0)
    
    # final steps
    PreFinal = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible")
    newImage.add_layer (PreFinal, 0)
    PreFinal.mode = LAYER_MODE_NORMAL   # default for normal luma
    PreFinal.opacity = 100.0            # default for normal luma
    PreFinal.name = "Preliminary Final to check luma "
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (PreFinal,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    #pdb.gimp_message (str(meanl))
    # WARNING: if the value is not integer python goes to the "else" branch
    # for instance having 120.1 python goes to the "else" branch 
    if int(meanl) in range (96,160):   # int is therefore mandatory for this range
        #pdb.gimp_message ("here?")                                 
        PreFinal.opacity = 90.0            # for normal luma
        pdb.gimp_drawable_curves_spline (PreFinal, HISTOGRAM_VALUE,
                                             6, (0.0,0.10, 0.5,0.55, 1.0,1.0)) 
    else:
        #pdb.gimp_message ("...or here?")                                 
        # in any other case
        PreFinal.opacity = 50.0
        if meanl < 64:
            pdb.gimp_drawable_curves_spline (PreFinal, HISTOGRAM_VALUE,
                                             6, (0.0,0.15, 0.5,0.7, 1.0,1.0)) 
        elif meanl < 96:
            pdb.gimp_drawable_curves_spline (PreFinal, HISTOGRAM_VALUE,
                                             6, (0.0,0.10, 0.5,0.6, 1.0,1.0)) 
        elif meanl > 192:
            pdb.gimp_drawable_curves_spline (PreFinal, HISTOGRAM_VALUE,
                                             6, (0.0,0.0, 0.7,0.5, 1.0,0.85))
        elif meanl > 160:
            pdb.gimp_drawable_curves_spline (PreFinal, HISTOGRAM_VALUE,
                                             6, (0.0,0.0, 0.6,0.5, 1.0,0.9))
    SCALE=1.25; COLORSPACE=0; 
    gegl_saturation (newImage, PreFinal, SCALE, COLORSPACE)
     
    plook=["small","medium","large"]
    option = option+" and edges="+plook[edgeslook]
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "outcome" 
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with effect B (MeanCurvature+Neon), option "+option
    
    if widthheight == 0: # size should be = input
        widthheight = inLayer.width+inLayer.height
    if widthheight == wrkwh: # size is already OK
        hiddenmsg = "OK"
    else:
        ratio = 1.00*widthheight/wrkwh  # float
        outW = int(outcome.width*ratio)
        outH = int(outcome.height*ratio)
        pdb.gimp_image_scale (newImage, outW, outH)
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="dinasset"
menu="<Image>/Diego/EasyEffect/"
copyrightYear="2021"
desc="Easy Effect Style B (MeanCurvature+Neon)"

register(
    "EasyEffectStyleB",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3600, (0, 6000, 100)),
        (PF_OPTION, "drawlook", "Draw look", 0, ("colour","B&W","mix")),
        (PF_OPTION, "curvature", "Mean curvature smooth", 1, ("small","medium","large")),
        (PF_OPTION, "texturelook", "Texture look", 1, ("delicate","normal","strong")),
        (PF_TOGGLE, "flatten", "Flatten Image?", True),
    ],
    [],
    EasyEffectStyleB,
    menu=menu
)

main()
